use core::{fmt::Write, any::Any};

use defmt;

use crate::uart::LitexUart;
use core::arch::asm;

#[defmt::global_logger]
struct DefmtLogger;

static mut LOGGER_SOCKET: Option<&'static mut smoltcp::socket::tcp::Socket> = None;

unsafe impl defmt::Logger for DefmtLogger {
    fn acquire() {
        // Sync methods left empty because we don't use any interrupts
    }

    unsafe fn flush() {
        // Sync methods left empty because we don't use any interrupts
    }

    unsafe fn release() {
        // Sync methods left empty because we don't use any interrupts
    }

    unsafe fn write(bytes: &[u8]) {
        static mut UART: Option<LitexUart> = None;
        if UART.is_none() {
            UART = Some(LitexUart::new(0xf000_4000));
        }


        let mut dev = UART.unwrap();
        for byte in bytes {
            while let Err(_) = dev.try_put_char(*byte) {}
        }
    }
}

/// Set the TCP socket to use for defmt logging.
///
/// SAFETY: Upcasts the reference from immutable ref to 'static mutable ref,
/// ensure this socket exists forever, or at least until you call this with a
/// None.
pub unsafe fn set_logger_socket(socket: Option<&smoltcp::socket::tcp::Socket>) {
    LOGGER_SOCKET = core::mem::transmute(socket);
}
